from datetime import datetime
from typing import Literal
from uuid import UUID

import pandas as pd
from pydantic import Field, field_serializer

from qnexus.models.annotations import Annotations
from qnexus.models.references.base import BaseRef


class ProjectRef(BaseRef):
    """Proxy object to a Project in Nexus."""

    annotations: Annotations
    contents_modified: datetime
    archived: bool = Field(default=False)
    id: UUID
    type: Literal["ProjectRef"] = "ProjectRef"

    @field_serializer("contents_modified")
    def serialize_modified(self, contents_modified: datetime | None) -> str | None:
        """Custom serializer for datetimes."""
        if contents_modified:
            return str(contents_modified)
        return None

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.annotations.name,
                "description": self.annotations.description,
                "created": self.annotations.created,
                "modified": self.annotations.modified,
                "contents_modified": self.contents_modified,
                "archived": self.archived,
                "id": self.id,
            },
            index=[0],
        )



"""Client API for HUGR in Nexus.

N.B. Nexus support for HUGR is experimental, and any HUGRs programs
uploaded to Nexus before stability is achieved might not work in the future.
"""

import base64
import warnings
from datetime import datetime
from typing import Any, Literal, Union, cast
from uuid import UUID

from hugr.hugr import Hugr
from hugr.ops import Module
from hugr.package import Package, PackagePointer

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
    merge_scope_from_context,
)
from qnexus.models import QuantinuumConfig
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    NameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import (
    DataframableList,
    ExecutionProgram,
    HUGRRef,
    ProjectRef,
)
from qnexus.models.scope import ScopeFilterEnum


class Params(
    SortFilter,
    PaginationFilter,
    NameFilter,
    CreatorFilter,
    ProjectRefFilter,
    PropertiesFilter,
    TimeFilter,
    ScopeFilter,
):
    """Params for filtering HUGRs."""


@merge_scope_from_context
@merge_project_from_context
def get_all(
    *,
    name_like: str | None = None,
    name_exact: list[str] | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> NexusIterator[HUGRRef]:
    """Get a NexusIterator over HUGRs with optional filters."""

    params = Params(
        name_like=name_like,
        name_exact=name_exact,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="HUGR",
        nexus_url="/api/hugr/v1beta",
        params=params,
        wrapper_method=_to_hugr_ref,
        nexus_client=get_nexus_client(),
    )


def _to_hugr_ref(page_json: dict[str, Any]) -> DataframableList[HUGRRef]:
    """Convert JSON response dict to a list of HUGRRefs."""

    hugr_refs: DataframableList[HUGRRef] = DataframableList([])

    for hugr_data in page_json["data"]:
        project_id = hugr_data["relationships"]["project"]["data"]["id"]
        project_details = next(
            proj for proj in page_json["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        hugr_refs.append(
            HUGRRef(
                id=UUID(hugr_data["id"]),
                annotations=Annotations.from_dict(hugr_data["attributes"]),
                project=project,
            )
        )
    return hugr_refs


@merge_scope_from_context
def get(
    *,
    id: Union[UUID, str, None] = None,
    name: str | None = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> HUGRRef:
    """
    Get a single HUGR using filters. Throws an exception if the filters do
    not match exactly one object.
    """

    if id:
        return _fetch_by_id(hugr_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        name_exact=[name] if name else None,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_properties_from_context
def upload(
    hugr_package: Package | PackagePointer | Hugr[Module],
    name: str,
    project: ProjectRef | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> HUGRRef:
    """Upload a HUGR to Nexus.

    N.B. HUGR support in Nexus is subject to change. Until full support is achieved any
    programs uploaded may not work in the future.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    match hugr_package:
        case PackagePointer():
            package = hugr_package.package
        case Hugr():
            package = Package([hugr_package])
        case Package():
            package = hugr_package
        case _:
            raise ValueError(f"Unsupported type for HUGR upload: {type(hugr_package)}.")
    attributes = {"contents": _encode_hugr(package)}

    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes.update(annotations)
    relationships = {"project": {"data": {"id": str(project.id), "type": "project"}}}

    req_dict = {
        "data": {
            "attributes": attributes,
            "relationships": relationships,
            "type": "hugr",
        }
    }

    res = get_nexus_client().post("/api/hugr/v1beta", json=req_dict)

    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return HUGRRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        project=project,
    )


@merge_properties_from_context
def update(
    ref: HUGRRef,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> HUGRRef:
    """Update the annotations on a HUGRRef."""
    ref_annotations = ref.annotations.model_dump()
    annotations = Annotations(
        name=name,
        description=description,
        properties=properties if properties else PropertiesDict(),
    ).model_dump(exclude_none=True)
    ref_annotations.update(annotations)

    req_dict = {
        "data": {
            "attributes": annotations,
            "relationships": {},
            "type": "hugr",
        }
    }

    res = get_nexus_client().patch(f"/api/hugr/v1beta/{ref.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_dict = res.json()["data"]

    return HUGRRef(
        id=UUID(res_dict["id"]),
        annotations=Annotations.from_dict(res_dict["attributes"]),
        project=ref.project,
    )


def cost(
    programs: HUGRRef | list[HUGRRef],
    n_shots: int | list[int],
    project: ProjectRef | None = None,
    system_name: Literal["Helios-1"] = "Helios-1",
) -> float:
    """Estimate the cost (in HQC) of running Hugr programs for n_shots
    number of shots on a Quantinuum Helios system.

    NB: This will execute a costing job on a dedicated cost estimation device.
        Once run, the cost will be visible also in the Nexus web portal
        as part of the job.
    """
    import qnexus as qnx

    warnings.warn(
        "hugr.cost() is deprecated. Please update to use hugr.cost_confidence() instead.",
        category=DeprecationWarning,
    )

    if isinstance(programs, HUGRRef):
        programs = [programs]

    job_ref = qnx.start_execute_job(
        programs=cast(list[ExecutionProgram], programs),
        n_shots=n_shots,
        backend_config=QuantinuumConfig(device_name=f"{system_name}SC"),
        project=project,
        name="Hugr cost estimation job",
    )
    status = qnx.jobs.wait_for(job_ref)

    return cast(float, status.cost)


def cost_confidence(
    programs: HUGRRef | list[HUGRRef],
    n_shots: int | list[int],
    project: ProjectRef | None = None,
    system_name: Literal["Helios-1"] = "Helios-1",
) -> list[tuple[float, float]]:
    """Estimate the cost (in HQC) of running Hugr programs for n_shots
    number of shots on a Quantinuum Helios system.

    Returns a list of tuples of (cost, confidence) for each job item.

    NB: This will execute a costing job on a dedicated cost estimation device.
        Once run, the cost will be visible also in the Nexus web portal
        as part of the job.
    """
    import qnexus as qnx

    if isinstance(programs, HUGRRef):
        programs = [programs]

    job_ref = qnx.start_execute_job(
        programs=cast(list[ExecutionProgram], programs),
        n_shots=n_shots,
        # No other parameters matter for cost estimation, so construct a minimal costing config
        backend_config=QuantinuumConfig(device_name=f"{system_name}SC"),
        project=project,
        name="Circuit cost confidence estimation job",
    )

    qnx.jobs.wait_for(job_ref)

    return qnx.jobs.cost_confidence(job_ref)


@merge_scope_from_context
def _fetch_by_id(
    hugr_id: UUID | str, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> HUGRRef:
    """Utility method for fetching directly by a unique identifier."""

    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/hugr/v1beta/{hugr_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    project_id = res_dict["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in res_dict["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    return HUGRRef(
        id=UUID(res_dict["data"]["id"]),
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        project=project,
    )


@merge_scope_from_context
def _fetch_hugr_package(
    handle: HUGRRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> Package:
    """Utility method for fetching a HUGR Package from a HUGRRef."""

    hugr_bytes = _fetch_hugr_bytes(handle=handle, scope=scope)
    return Package.from_bytes(envelope=hugr_bytes)


@merge_scope_from_context
def _fetch_hugr_bytes(
    handle: HUGRRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> bytes:
    """Utility method for fetching HUGR bytes from a HUGRRef."""
    res = get_nexus_client().get(
        f"/api/hugr/v1beta/{handle.id}",
        params={"scope": scope.value},
    )
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    contents = res.json()["data"]["attributes"]["contents"]
    return base64.b64decode(contents)


def _encode_hugr(hugr_package: Package) -> str:
    """Utility method for encoding a HUGR Package as base64-encoded string"""
    return base64.b64encode(hugr_package.to_bytes()).decode("utf-8")



"""Client API for teams in Nexus."""

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.context import merge_scope_from_context
from qnexus.models.references import DataframableList, TeamRef
from qnexus.models.scope import ScopeFilterEnum


@merge_scope_from_context
def get_all(scope: ScopeFilterEnum = ScopeFilterEnum.USER) -> DataframableList[TeamRef]:
    """No fuzzy name matching."""
    res = get_nexus_client().get(
        "/api/teams/v1beta2",
        params={"scope": scope.value},
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    return DataframableList(
        [
            TeamRef(
                id=team["id"],
                name=team["attributes"]["name"],
                description=team["attributes"]["description"],
            )
            for team in res.json()["data"]
        ]
    )


@merge_scope_from_context
def get(name: str, scope: ScopeFilterEnum = ScopeFilterEnum.USER) -> TeamRef:
    """
    Get a single team using filters. Throws an exception if the filters do not
    match exactly one object.
    """
    res = get_nexus_client().get(
        "/api/teams/v1beta2",
        params={
            "filter[team][name]": name,
            "scope": scope.value,
        },
    )

    if res.status_code == 404 or res.json()["data"] == []:
        raise qnx_exc.ZeroMatches

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    teams_list = [
        TeamRef(
            id=team["id"],
            name=team["attributes"]["name"],
            description=team["attributes"]["description"],
        )
        for team in res.json()["data"]
    ]

    if len(teams_list) > 1:
        print(teams_list)
        raise qnx_exc.NoUniqueMatch

    return teams_list[0]


@merge_scope_from_context
def _fetch_by_id(
    team_id: str, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> TeamRef:
    """
    Get a single team by id.
    """
    res = get_nexus_client().get(
        f"/api/teams/v1beta2/{team_id}",
        params={"scope": scope.value},
    )

    if res.status_code == 404:
        raise qnx_exc.ZeroMatches

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    team_dict = res.json()["data"]

    return TeamRef(
        id=team_dict["id"],
        name=team_dict["attributes"]["name"],
        description=team_dict["attributes"]["description"],
    )


def create(name: str, description: str | None = None) -> TeamRef:
    """Create a team in Nexus."""

    resp = get_nexus_client().post(
        "/api/teams/v1beta2",
        json={
            "data": {
                "attributes": {
                    "name": name,
                    "description": description,
                    "display_name": name,
                },
                "relationships": {},
                "type": "team",
            },
        },
    )

    if resp.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=resp.text, status_code=resp.status_code
        )

    team_dict = resp.json()["data"]
    return TeamRef(
        id=team_dict["id"],
        name=team_dict["attributes"]["name"],
        description=team_dict["attributes"]["description"],
    )



"""Tests related to running jobs."""

import uuid
from datetime import datetime

import pytest
from quantinuum_schemas.models.backend_config import SeleneConfig

import qnexus as qnx
from qnexus.models.annotations import Annotations
from qnexus.models.references import (
    HUGRRef,
    ProjectRef,
)


@pytest.mark.parametrize(
    "n_shots, max_cost, n_qubits, error_param",
    [
        ([10], [10, 20], [5, 10], "n_shots"),
        ([10, 20], [10], [5, 10], "max_cost"),
        ([10, 20], [10, 20], [5], "n_qubits"),
    ],
)
def test_job_parameterization(
    n_shots: list[int],
    max_cost: list[float],
    n_qubits: list[int],
    error_param: str,
) -> None:
    """Test that we can parameterize max_cost and n_qubits for a Hugr program submission."""

    my_proj = ProjectRef(
        id=uuid.uuid4(), annotations=Annotations(), contents_modified=datetime.now()
    )

    hugr_ref = HUGRRef(
        id=uuid.uuid4(),
        annotations=Annotations(),
        project=my_proj,
    )

    with pytest.raises(
        ValueError, match=f"Number of programs must equal number of {error_param}."
    ):
        qnx.start_execute_job(
            programs=[hugr_ref, hugr_ref],
            n_shots=n_shots,
            backend_config=SeleneConfig(),
            project=my_proj,
            name="This job will never run",
            max_cost=max_cost,
            n_qubits=n_qubits,
        )



#!/usr/bin/env python3

import re
from pathlib import Path

import requests


def extract_dependencies(pyproject_path):
    """Extract dependencies from the dependencies block in pyproject.toml."""
    deps = []
    in_deps_block = False
    with open(pyproject_path, "r") as f:
        for line in f:
            if line.strip().startswith("dependencies = ["):
                in_deps_block = True
                continue
            if in_deps_block:
                if line.strip().startswith("]"):
                    break
                dep = line.strip().strip('",')
                if dep:
                    pkg_name = re.split(r"[ <>=]", dep)[0]
                    deps.append((pkg_name, dep))
    return deps


def get_pypi_info(pkg_name):
    """Fetch package info from PyPI."""
    try:
        resp = requests.get(f"https://pypi.org/pypi/{pkg_name}/json", timeout=5)
        if resp.status_code == 200:
            info = resp.json()["info"]
            summary = info.get("summary", "").strip()
            homepage = (
                info.get("home_page", None)
                or info.get("project_urls", None).get("homepage", None)
                or info.get("project_url", None)
            )
            if homepage:
                homepage = homepage.strip()
            if not homepage:
                homepage = f"https://pypi.org/project/{pkg_name}/"
            return summary, homepage
    except Exception:
        pass
    return "No description found.", f"https://pypi.org/project/{pkg_name}/"


def write_markdown_table(deps, md_path):
    """Write dependencies and descriptions as a markdown table."""
    with open(md_path, "w") as f:
        f.write("# Project Dependencies\n\n")
        f.write("| Package | Version Spec | Description | Homepage |\n")
        f.write("|---------|--------------|-------------|----------|\n")
        for pkg_name, dep in deps:
            desc, homepage = get_pypi_info(pkg_name)
            version_spec = dep[len(pkg_name) :].strip()
            f.write(
                f"| `{pkg_name}` | `{version_spec}` | {desc} | [{homepage}]({homepage}) |\n"
            )


if __name__ == "__main__":
    pyproject_file = Path(__file__).parent.parent / "pyproject.toml"
    md_file = Path(__file__).parent.parent / "DEPENDENCIES.md"
    dependencies = extract_dependencies(pyproject_file)
    write_markdown_table(dependencies, md_file)
    print(f"Wrote {len(dependencies)} dependencies as a markdown table to {md_file}")



"""Custom exceptions thrown in qnexus usage."""

import json
from typing import Optional


class AuthenticationError(Exception):
    "Raised when there is an issue authenticating with the Nexus API."


class ResourceFetchFailed(Exception):
    """
    Raised when a resource cannot be fetched from the platform to be read, or
    when expected data wasn't found on an object.
    """

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = f"Failed to fetch resource with status code: {self.status_code}\n"
        try:
            msg_data = json.loads(message)
            self.err += f"Message: {msg_data['message']}"
            for fld, val in msg_data.items():
                if fld != "message":
                    self.err += f"\n{fld}: {val}"
        except json.JSONDecodeError:
            self.err += f"Message: {message}"
        super().__init__(self.err)


class ZeroMatches(Exception):
    """Raised when a ``get`` call finds no matches."""


class NoUniqueMatch(Exception):
    """Raised when a ``get`` call finds more than one match."""


class JobError(Exception):
    """Raised when a Nexus Job has errored."""


class ResourceCreateFailed(Exception):
    """Raised when a resource couldn't be created on the platform."""

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to create resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class ResourceDeleteFailed(Exception):
    """Raised when a resource couldn't be deleted on the platform."""

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to delete resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class ResourceUpdateFailed(Exception):
    """Raised when a resource cannot be updated on the platform."""

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to update resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class IncompatibleResultVersion(Exception):
    """Raised when the user tries to get results in a version that's not possible"""



"""Test submitting compiler options for compilation jobs."""

from typing import Callable, ContextManager

import pytest
from pytket.circuit import Circuit, OpType

import qnexus as qnx
from qnexus.models.references import (
    CompilationResultRef,
    CompileJobRef,
)


def assert_op_count_in_circuit(
    circuit: Circuit, op: OpType, should_be_gt_0: bool
) -> None:
    """Assert that the number of `op` gates in the `circuit` is 0 or >0."""
    total_ops = sum([1 for circ_op in circuit.get_commands() if circ_op.op.type == op])
    if should_be_gt_0:
        assert total_ops > 0, (
            f"Expected >0 {op} gates in the circuit, but found {total_ops}."
        )
    else:
        assert total_ops == 0, (
            f"Expected 0 {op} gates in the circuit, but found {total_ops}."
        )


@pytest.mark.parametrize(
    "target_2qb_gate, expect_tk2_gt_0, expect_zzphase_gt_0, expect_zzmax_gt_0",
    [
        ("TK2", True, False, False),
        ("ZZPhase", False, True, False),
        ("ZZMax", False, False, True),
    ],
    ids=["TK2", "ZZPhase", "ZZMax"],
)
def test_target_2qb_gate(
    create_compile_job_in_project: Callable[..., ContextManager[CompileJobRef]],
    test_suite_name: str,
    test_case_name: str,
    test_qv_circuit: Circuit,
    target_2qb_gate: str,
    expect_tk2_gt_0: bool,
    expect_zzphase_gt_0: bool,
    expect_zzmax_gt_0: bool,
) -> None:
    """Test that we can submit compile jobs with different `target_2qb_gate`."""

    project_name = f"project for {test_suite_name}"
    circuit_name = f"circuit foro {test_case_name}"
    compile_job_name = f"compile job for {test_case_name}"

    device_name = "H2-Emulator"
    config = qnx.QuantinuumConfig(
        device_name=device_name,
        compiler_options={"target_2qb_gate": target_2qb_gate},
    )

    with create_compile_job_in_project(
        project_name=project_name,
        job_name=compile_job_name,
        circuit=test_qv_circuit,
        circuit_name=circuit_name,
        backend_config=config,
        skip_intermediate_circuits=True,
    ) as compile_job_ref:
        qnx.jobs.wait_for(compile_job_ref)
        compile_result = qnx.jobs.results(compile_job_ref)[0]
        assert isinstance(compile_result, CompilationResultRef)
        compiled_circuit_ref = compile_result.get_output()
        compiled_circuit = compiled_circuit_ref.download_circuit()

        assert_op_count_in_circuit(compiled_circuit, OpType.TK2, expect_tk2_gt_0)
        assert_op_count_in_circuit(
            compiled_circuit, OpType.ZZPhase, expect_zzphase_gt_0
        )
        assert_op_count_in_circuit(compiled_circuit, OpType.ZZMax, expect_zzmax_gt_0)



"""Simple guppylang program from https://docs.quantinuum.com/guppy"""

from pathlib import Path
from sys import argv

from guppylang import guppy
from guppylang.std.builtins import result
from guppylang.std.quantum import cx, h, measure, qubit, x, z


@guppy
def bell() -> tuple[qubit, qubit]:
    """Constructs a bell state."""
    q1, q2 = qubit(), qubit()
    h(q1)
    cx(q1, q2)
    return q1, q2


@guppy
def main() -> None:
    src = qubit()
    x(src)
    alice, bob = bell()

    cx(src, alice)
    h(src)
    if measure(alice):
        x(bob)
    if measure(src):
        z(bob)

    result("teleported", measure(bob))


main.check()
program = main.compile_function()
# remove the .gpy suffix and add .hugr
Path(argv[0]).with_suffix("").with_suffix(".hugr").write_bytes(program.to_bytes())



"""Client API for circuits in Nexus."""

from datetime import datetime
from typing import Any, Union, cast
from uuid import UUID
from warnings import warn

from pytket.circuit import Circuit
from pytket.utils.serialization.migration import circuit_dict_from_pytket1_dict
from quantinuum_schemas.models.backend_config import BackendConfig, QuantinuumConfig

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
    merge_scope_from_context,
)
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    NameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import (
    CircuitRef,
    DataframableList,
    ExecutionProgram,
    ProjectRef,
)
from qnexus.models.scope import ScopeFilterEnum


class Params(
    SortFilter,
    PaginationFilter,
    NameFilter,
    CreatorFilter,
    ProjectRefFilter,
    PropertiesFilter,
    TimeFilter,
    ScopeFilter,
):
    """Params for filtering circuits."""


@merge_scope_from_context
@merge_project_from_context
def get_all(
    *,
    name_like: str | None = None,
    name_exact: list[str] | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> NexusIterator[CircuitRef]:
    """Get a NexusIterator over circuits with optional filters."""

    params = Params(
        name_like=name_like,
        name_exact=name_exact,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="Circuit",
        nexus_url="/api/circuits/v1beta2",
        params=params,
        wrapper_method=_to_circuitref,
        nexus_client=get_nexus_client(),
    )


def _to_circuitref(page_json: dict[str, Any]) -> DataframableList[CircuitRef]:
    """Convert JSON response dict to a list of CircuitRefs."""

    circuit_refs: DataframableList[CircuitRef] = DataframableList([])

    for circuit_data in page_json["data"]:
        project_id = circuit_data["relationships"]["project"]["data"]["id"]
        project_details = next(
            proj for proj in page_json["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        circuit_refs.append(
            CircuitRef(
                id=UUID(circuit_data["id"]),
                annotations=Annotations.from_dict(circuit_data["attributes"]),
                project=project,
            )
        )
    return circuit_refs


@merge_scope_from_context
def get(
    *,
    id: Union[UUID, str, None] = None,
    name: str | None = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> CircuitRef:
    """
    Get a single circuit using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(circuit_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        name_exact=[name] if name else None,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_properties_from_context
def upload(
    circuit: Circuit,
    project: ProjectRef | None = None,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> CircuitRef:
    """Upload a pytket Circuit to Nexus."""
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    circuit_dict = circuit.to_dict()
    circuit_name = name if name else circuit.name
    if circuit_name is None:
        raise ValueError("Circuit must have a name to be uploaded")

    annotations = CreateAnnotations(
        name=circuit_name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    circuit_dict.update(annotations)
    relationships = {"project": {"data": {"id": str(project.id), "type": "project"}}}

    req_dict = {
        "data": {
            "attributes": circuit_dict,
            "relationships": relationships,
            "type": "circuit",
        }
    }

    res = get_nexus_client().post("/api/circuits/v1beta2", json=req_dict)

    # https://cqc.atlassian.net/browse/MUS-3054
    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return CircuitRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        project=project,
    )


@merge_properties_from_context
def update(
    ref: CircuitRef,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> CircuitRef:
    """Update the annotations on a CircuitRef."""
    ref_annotations = ref.annotations.model_dump()
    annotations = Annotations(
        name=name,
        description=description,
        properties=properties if properties else PropertiesDict(),
    ).model_dump(exclude_none=True)
    ref_annotations.update(annotations)

    req_dict = {
        "data": {
            "attributes": annotations,
            "relationships": {},
            "type": "circuit",
        }
    }

    res = get_nexus_client().patch(f"/api/circuits/v1beta2/{ref.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_dict = res.json()["data"]

    return CircuitRef(
        id=UUID(res_dict["id"]),
        annotations=Annotations.from_dict(res_dict["attributes"]),
        project=ref.project,
    )


@merge_scope_from_context
def _fetch_by_id(
    circuit_id: UUID | str, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> CircuitRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/circuits/v1beta2/{circuit_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    project_id = res_dict["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in res_dict["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    return CircuitRef(
        id=UUID(res_dict["data"]["id"]),
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        project=project,
    )


@merge_scope_from_context
def _fetch_circuit(
    handle: CircuitRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> Circuit:
    """Utility method for fetching a pytket circuit from a CircuitRef."""
    res = get_nexus_client().get(
        f"/api/circuits/v1beta2/{handle.id}",
        params={"scope": scope.value},
    )
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_data_attributes_dict = res.json()["data"]["attributes"]
    circuit_dict = {k: v for k, v in res_data_attributes_dict.items() if v is not None}

    return Circuit.from_dict(circuit_dict_from_pytket1_dict(circuit_dict))


def cost(
    circuit_ref: CircuitRef | list[CircuitRef],
    n_shots: int | list[int],
    backend_config: BackendConfig,
    syntax_checker: str | None = None,
    project: ProjectRef | None = None,
) -> float | None:
    """Estimate the cost (in HQC) of running Circuit programs for n_shots
    number of shots on a Quantinuum H2 system.

    NB: This will execute a costing job on a dedicated cost estimation device.
        Once run, the cost will be visible also in the Nexus web portal
        as part of the job.

        If a project is not provided, it will be taken from either the active context
        or the ProjectRef listed on the first CircuitRef.

        Future versions of this function will require a ProjectRef to be provided.
    """

    import qnexus as qnx

    if not isinstance(backend_config, QuantinuumConfig):
        raise ValueError(
            "QuantinuumConfig is the only supported backend config for circuit cost estimation."
        )

    programs = circuit_ref
    if isinstance(programs, CircuitRef):
        programs = [programs]

    project = project or qnx.context.get_active_project(project_required=False)
    if project is None:
        warn(
            "No ProjectRef was provided in function arguments. "
            "Taking ProjectRef from the first CircuitRef. "
            "In future qnexus versions, a ProjectRef will be required.",
            DeprecationWarning,
        )
        project = programs[0].project

    syntax_checker_device_name = backend_config.device_name

    if syntax_checker is not None:
        syntax_checker_device_name = syntax_checker

    if not syntax_checker_device_name.startswith("H2-"):
        raise ValueError("Circuit cost estimation is only supported for H2-x systems.")

    if not syntax_checker_device_name.endswith("SC"):
        syntax_checker_device_name.strip(
            "E"
        )  # costing for H2-xE devices is the same as for H2-x
        syntax_checker_device_name += "SC"

    job_ref = qnx.start_execute_job(
        programs=cast(list[ExecutionProgram], programs),
        n_shots=n_shots,
        # No other parameters matter for cost estimation, so construct a minimal costing config
        backend_config=QuantinuumConfig(device_name=syntax_checker_device_name),
        project=project,
        name="Circuit cost estimation job",
    )
    status = qnx.jobs.wait_for(job_ref)

    return cast(float, status.cost)



"""Enum for scope. Separate file to avoid circular imports."""

from enum import Enum


class ScopeFilterEnum(str, Enum):
    """ScopeFilterEnum model."""

    USER = "user"
    ORG_ADMIN = "org_admin"
    GLOBAL_ADMIN = "global_admin"
    HIGHEST = "highest"



"""Client API for execution in Nexus."""

from typing import Union, cast

from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult

import qnexus.exceptions as qnx_exc
from qnexus.client import circuits as circuit_api
from qnexus.client import get_nexus_client
from qnexus.client import hugr as hugr_api
from qnexus.client import qir as qir_api
from qnexus.client.utils import accept_circuits_for_programs
from qnexus.context import (
    get_active_project,
    merge_properties_from_context,
    merge_scope_from_context,
)
from qnexus.models import BackendConfig, StoredBackendInfo, to_pytket_backend_info
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.job_status import JobStatus, JobStatusEnum
from qnexus.models.language import Language
from qnexus.models.references import (
    CircuitRef,
    DataframableList,
    ExecuteJobRef,
    ExecutionProgram,
    ExecutionResultRef,
    GpuDecoderConfigRef,
    HUGRRef,
    IncompleteJobItemRef,
    JobType,
    ProjectRef,
    QIRRef,
    QIRResult,
    ResultType,
    ResultVersions,
    WasmModuleRef,
)
from qnexus.models.scope import ScopeFilterEnum
from qnexus.models.utils import assert_never, truncate_to_2dp


@accept_circuits_for_programs
@merge_properties_from_context
def start_execute_job(
    programs: ExecutionProgram | list[ExecutionProgram],
    n_shots: int | list[int] | list[None],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    properties: PropertiesDict | None = None,
    project: ProjectRef | None = None,
    valid_check: bool = True,
    language: Language = Language.AUTO,
    credential_name: str | None = None,
    wasm_module: WasmModuleRef | None = None,
    gpu_decoder_config: GpuDecoderConfigRef | None = None,
    user_group: str | None = None,
    max_cost: float | list[float] | list[None] = list(),
    n_qubits: int | list[int] | list[None] = list(),
) -> ExecuteJobRef:
    """
    Submit an execute job to be run in Nexus. Returns an ``ExecuteJobRef``
    object which can be used to check the job's status.  See ``qnexus.execute``
    for a utility method that waits for the results and returns them.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    program_ids = (
        [str(p.id) for p in programs]
        if isinstance(programs, list)
        else [str(programs.id)]
    )

    if isinstance(n_shots, int):
        n_shots = [n_shots] * len(program_ids)
    if isinstance(max_cost, (int, float)):
        max_cost = [float(max_cost)] * len(program_ids)
    elif not max_cost:
        max_cost = [None] * len(program_ids)
    if isinstance(n_qubits, int):
        n_qubits = [n_qubits] * len(program_ids)
    elif not n_qubits:
        n_qubits = [None] * len(program_ids)

    if len(n_shots) != len(program_ids):
        raise ValueError("Number of programs must equal number of n_shots.")
    if len(max_cost) != len(program_ids):
        raise ValueError("Number of programs must equal number of max_cost.")
    if len(n_qubits) != len(program_ids):
        raise ValueError("Number of programs must equal number of n_qubits.")

    attributes_dict = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)

    items = [
        {
            "program_id": program_id,
            "n_shots": n_shot,
            **({"max_cost": mc} if mc is not None else {}),
            **({"n_qubits": nq} if nq is not None else {}),
        }
        for program_id, n_shot, mc, nq in zip(program_ids, n_shots, max_cost, n_qubits)
    ]

    attributes_dict.update(
        {
            "job_type": "execute",
            "definition": {
                "job_definition_type": "execute_job_definition",
                "backend_config": backend_config.model_dump(exclude_none=True),
                "user_group": user_group,
                "valid_check": valid_check,
                "language": (
                    language.value if isinstance(language, Language) else language
                ),
                "wasm_module_id": str(wasm_module.id) if wasm_module else None,
                "gpu_decoder_config_id": (
                    str(gpu_decoder_config.id) if gpu_decoder_config else None
                ),
                "credential_name": credential_name,
                "items": items,
            },
        }
    )
    relationships = {
        "project": {"data": {"id": str(project.id), "type": "project"}},
    }
    req_dict = {
        "data": {
            "attributes": attributes_dict,
            "relationships": relationships,
            "type": "job",
        }
    }

    resp = get_nexus_client().post(
        "/api/jobs/v1beta3",
        json=req_dict,
    )
    if resp.status_code != 202:
        raise qnx_exc.ResourceCreateFailed(
            message=resp.text, status_code=resp.status_code
        )

    return ExecuteJobRef(
        id=resp.json()["data"]["id"],
        annotations=Annotations.from_dict(resp.json()["data"]["attributes"]),
        job_type=JobType.EXECUTE,
        last_status=JobStatusEnum.SUBMITTED,
        last_message="",
        last_status_detail=JobStatus.from_dict(
            resp.json()["data"]["attributes"]["status"]
        ),
        project=project,
        backend_config_store=backend_config,
    )


@merge_scope_from_context
def _results(
    execute_job: ExecuteJobRef,
    allow_incomplete: bool = True,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> DataframableList[ExecutionResultRef | IncompleteJobItemRef]:
    """Get the results from an execute job."""

    resp = get_nexus_client().get(
        f"/api/jobs/v1beta3/{execute_job.id}",
        params={"scope": scope.value},
    )
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    resp_data = resp.json()["data"]
    job_status = resp_data["attributes"]["status"]["status"]

    if job_status != "COMPLETED" and allow_incomplete is not True:
        raise qnx_exc.ResourceFetchFailed(message=f"Job status: {job_status}")

    execute_results: DataframableList[ExecutionResultRef | IncompleteJobItemRef] = (
        DataframableList([])
    )

    program_types = resp_data["relationships"]["programs"]["data"]
    program_type_dict = {p["id"]: p["type"] for p in program_types}

    for item in resp_data["attributes"]["definition"]["items"]:
        result_type: ResultType | None = None

        match item.get("result_type", None):
            case ResultType.QSYS:
                result_type = ResultType.QSYS
            case ResultType.PYTKET:
                result_type = ResultType.PYTKET
            case None:
                result_type = None
            case _:
                assert_never(item["result_type"])

        # Check if item is in a state that returns results
        # and has results
        if (
            item["status"]["status"]
            in (
                "CANCELLED",
                "ERROR",
                "DEPLETED",
                "TERMINATED",
                "COMPLETED",
                "RUNNING",
                "QUEUED",
            )
            and result_type
        ):
            result_ref = ExecutionResultRef(
                id=item["result_id"],
                job_item_id=item.get("external_handle", None),
                job_item_integer_id=item.get("item_id", None),
                annotations=execute_job.annotations,
                project=execute_job.project,
                result_type=result_type,
                cost=truncate_to_2dp(item["status"].get("cost", None)),
                last_status_detail=JobStatus.from_dict(item["status"]),
            )

            execute_results.append(result_ref)

        elif allow_incomplete is True:
            # Job item is not complete, return an IncompleteJobItemRef
            program_id = item.get("program_id", None)
            program_type = program_type_dict[program_id]

            incomplete_ref = IncompleteJobItemRef(
                job_item_id=item.get("external_handle", None),
                job_item_integer_id=item.get("item_id", None),
                annotations=execute_job.annotations,
                project=execute_job.project,
                program_type=program_type,
                program_id=program_id,
                job_type=JobType.EXECUTE,
                last_status=JobStatusEnum[item["status"]["status"]],
                last_message=item["status"].get("message", ""),
                last_status_detail=JobStatus.from_dict(item["status"]),
            )
            execute_results.append(incomplete_ref)

    return execute_results


@merge_scope_from_context
def _fetch_pytket_execution_result(
    result_ref: ExecutionResultRef,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> tuple[BackendResult, BackendInfo, Union[CircuitRef, QIRRef]]:
    """Get the results for an execute job item."""
    assert result_ref.result_type == ResultType.PYTKET, "Incorrect result type"

    res = get_nexus_client().get(
        f"/api/results/v1beta3/{result_ref.id}",
        params={"scope": scope.value},
    )
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_dict = res.json()
    program_data = res_dict["data"]["relationships"]["program"]["data"]
    program_id = program_data["id"]
    program_type = program_data["type"]

    input_program: CircuitRef | QIRRef
    match program_type:
        case "circuit":
            input_program = circuit_api._fetch_by_id(program_id)
        case "qir":
            input_program = qir_api._fetch_by_id(program_id)
        case _:
            raise ValueError(f"Unknown program type {type}")

    results_data = res_dict["data"]["attributes"]

    results_dict = {k: v for k, v in results_data.items() if v != [] and v is not None}

    backend_result = BackendResult.from_dict(results_dict)

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = to_pytket_backend_info(
        StoredBackendInfo(**backend_info_data["attributes"])
    )

    return (backend_result, backend_info, input_program)


@merge_scope_from_context
def _fetch_qsys_execution_result(
    result_ref: ExecutionResultRef,
    version: ResultVersions,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> tuple[QsysResult | QIRResult, BackendInfo, HUGRRef | QIRRef]:
    """Get the results of a next-gen Qsys execute job."""
    assert result_ref.result_type == ResultType.QSYS, "Incorrect result type"

    chunk_number = 0
    params = {
        "version": version.value,
        "chunk_number": chunk_number,
        "scope": scope.value,
    }

    res = get_nexus_client().get(
        f"/api/qsys_results/v1beta2/partial/{result_ref.id}", params=params
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    # This is only needed to be set once, as subsequent calls will
    # return the same information for the relationships.
    res_dict = res.json()
    input_program_id = res_dict["data"]["relationships"]["program"]["data"]["id"]

    input_program: HUGRRef | QIRRef
    result: QsysResult | QIRResult
    match res_dict["data"]["relationships"]["program"]["data"]["type"]:
        case "hugr":
            input_program = hugr_api._fetch_by_id(
                input_program_id,
            )
            result = QsysResult(res_dict["data"]["attributes"].get("results"))
        case "qir":
            input_program = qir_api._fetch_by_id(
                input_program_id,
            )
            if version == ResultVersions.DEFAULT:
                result = QIRResult(res_dict["data"]["attributes"].get("results"))
            else:
                result = QsysResult(res_dict["data"]["attributes"].get("results"))

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = to_pytket_backend_info(
        StoredBackendInfo(**backend_info_data["attributes"])
    )

    # We shouldn't be doing infinite loops, but the API currently doesn't
    # provide a way to know how many chunks there are, so we loop until we
    # get all of them.
    while True:
        chunk_number += 1
        params["chunk_number"] = chunk_number
        partial = get_nexus_client().get(
            f"/api/qsys_results/v1beta2/partial/{result_ref.id}", params=params
        )
        if partial.status_code == 404:
            # No more chunks. Stop here.
            break
        if partial.status_code != 200:
            raise qnx_exc.ResourceFetchFailed(
                message=res.text, status_code=partial.status_code
            )
        if isinstance(result.results, str):
            assert (
                version == ResultVersions.DEFAULT
            )  # Only QIR outputs are in this mode
            prev_str = result.results.split("END")[
                0
            ]  # remove the end tag from result.results
            next_str = "\n".join(
                [
                    line
                    for line in QIRResult(
                        partial.json()["data"]["attributes"]["results"]
                    ).results.splitlines()
                    if "OUTPUT" in line
                ]
            )  # just the output lines
            result.results += (
                prev_str + next_str + "END\t0\n"
            )  # join everything back up
        else:
            next_res = QsysResult(partial.json()["data"]["attributes"]["results"])
            result.results.extend(next_res.results)

    return (
        result,
        backend_info,
        input_program,
    )



"""Client API for jobs in Nexus."""

import abc
import asyncio
import json
import logging
import ssl
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Type, Union, cast, overload
from uuid import UUID

import httpx
from quantinuum_schemas.models.backend_config import config_name_to_class
from quantinuum_schemas.models.hypertket_config import HyperTketConfig
from websockets.asyncio.client import connect, process_exception
from websockets.exceptions import ConnectionClosed

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.jobs import _compile, _execute
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import accept_circuits_for_programs, handle_fetch_errors
from qnexus.config import CONFIG
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
    merge_scope_from_context,
)
from qnexus.models import BackendConfig
from qnexus.models.annotations import Annotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    JobStatusFilter,
    JobTypeFilter,
    NameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.job_status import WAITING_STATUS, JobStatus, JobStatusEnum
from qnexus.models.language import Language
from qnexus.models.references import (
    CircuitRef,
    CompilationResultRef,
    CompileJobRef,
    DataframableList,
    ExecuteJobRef,
    ExecutionProgram,
    ExecutionResult,
    ExecutionResultRef,
    GpuDecoderConfigRef,
    IncompleteJobItemRef,
    JobRef,
    JobType,
    ProjectRef,
    SystemRef,
    WasmModuleRef,
)
from qnexus.models.scope import ScopeFilterEnum
from qnexus.models.utils import assert_never

logger = logging.getLogger(__name__)

EPOCH_START = datetime(1970, 1, 1, tzinfo=timezone.utc)


class RemoteRetryStrategy(str, Enum):
    """Strategy to use when retrying jobs.

    Each strategy defines how the system should approach resolving
    potential conflicts with remote state.

    FULL_RESTART will act as though the job is entirely fresh and
    re-perform every action.
    """

    FULL_RESTART = "FULL_RESTART"


@dataclass
class WaitStrategy(abc.ABC):
    wait_for_status: JobStatusEnum = JobStatusEnum.COMPLETED

    @abc.abstractmethod
    async def get_status(self, job: JobRef) -> JobStatus:
        pass

    def _finished(self, job_status: JobStatus) -> bool:
        return (
            job_status.status not in WAITING_STATUS
            or job_status.status == self.wait_for_status
        )


@dataclass
class WebsocketStrategy(WaitStrategy):
    """Use a websocket connection for real-time updates.

    Best for short-running jobs (<10 minutes).
    """

    async def get_status(self, job: JobRef) -> JobStatus:
        """Check the Status of a Job via a websocket connection.
        Will use SSO tokens."""

        job_status = status(job)
        logger.debug("Job %s initial status: %s", job.id, job_status.status.value)
        if self._finished(job_status):
            return job_status

        ssl_context = httpx.create_ssl_context(verify=CONFIG.httpx_verify)

        def _process_exception(exc: Exception) -> Exception | None:
            """Utility wrapper around process_exception that tells the websockets
            library not to auto-retry SSLErrors as they are usually not recoverable.

            Unfortunately SSLError inherits from OSError which websockets will always
            retried when `connect` is used in an async for loop.
            """
            if isinstance(exc, ssl.SSLError):
                return exc
            return process_exception(exc)

        additional_headers = {
            # TODO, this cookie will expire frequently
            "Cookie": f"myqos_id={get_nexus_client().auth.cookies.get('myqos_id')}"  # type: ignore
        }
        logger.debug("Job %s: opening websocket connection", job.id)
        async for websocket in connect(
            f"{CONFIG.websockets_url}/api/jobs/v1beta3/{job.id}/attributes/status/ws",
            ssl=ssl_context,
            additional_headers=additional_headers,
            process_exception=_process_exception,
            logger=logger,
        ):
            try:
                async for status_json in websocket:
                    job_status = JobStatus.from_dict(json.loads(status_json))
                    logger.debug(
                        "Job %s websocket update: %s",
                        job.id,
                        job_status.status.value,
                    )

                    if self._finished(job_status):
                        break
                break
            except ConnectionClosed:
                logger.debug(
                    "Job %s: websocket connection closed, attempting to reconnect",
                    job.id,
                )
                continue
            finally:
                try:
                    await websocket.close(code=1000, reason="Client closed connection")
                except GeneratorExit:
                    pass

        return job_status


@dataclass
class PollingStrategy(WaitStrategy):
    """Use exponential backoff polling.

    More robust for long-running jobs (>10 minutes).

    Attributes:
        initial_interval: Starting poll interval in seconds.
        max_interval_queued: Maximum poll interval when job is queued.
        max_interval_running: Maximum poll interval when job is running/submitted.
        backoff_factor: Multiplier for interval after each poll.
    """

    initial_interval: float = 1.0
    max_interval_queued: float = 1200.0
    max_interval_running: float = 180.0
    backoff_factor: float = 2.0

    async def get_status(self, job: JobRef) -> JobStatus:
        """Poll job status with exponential backoff and adaptive intervals.

        Uses different maximum poll intervals based on job state:
        - QUEUED: Polls less frequently (default 20 min) since queue position changes slowly
        - RUNNING/SUBMITTED: Polls more frequently (default 3 min) for responsiveness

        Args:
            job: The job to monitor.
            wait_for_status: The status to wait for.
            strategy: Polling configuration.

        Returns:
            The final JobStatus when the target status is reached or job terminates.
        """
        interval = self.initial_interval
        logger.debug(
            "Starting polling for job %s (target: %s, interval: %.1fs, "
            "max queued: %.1fs, max running: %.1fs)",
            job.id,
            self.wait_for_status.value,
            self.initial_interval,
            self.max_interval_queued,
            self.max_interval_running,
        )

        while True:
            job_status = status(job)

            # Adapt max interval based on job state
            if job_status.status == JobStatusEnum.QUEUED:
                max_interval = self.max_interval_queued
            else:
                max_interval = self.max_interval_running

            # Clamp interval to current max (allows faster polling when transitioning
            # from QUEUED to RUNNING)
            interval = min(interval, max_interval)

            logger.debug(
                "Job %s status: %s (next poll in %.1fs, max: %.1fs)",
                job.id,
                job_status.status.value,
                interval,
                max_interval,
            )

            if self._finished(job_status):
                logger.debug(
                    "Job %s reached status: %s", job.id, job_status.status.value
                )
                return job_status

            await asyncio.sleep(interval)
            interval = min(interval * self.backoff_factor, max_interval)


@dataclass
class HybridStrategy(WebsocketStrategy, PollingStrategy):
    """Start with websocket, fall back to polling.

    Recommended for most use cases.

    Attributes:
        websocket_timeout: How long to use websocket before switching to polling.
    """

    websocket_timeout: float = 600.0

    async def get_status(self, job: JobRef) -> JobStatus:
        """Use websocket for initial period, then fall back to polling.

        Args:
            job: The job to monitor.
            wait_for_status: The status to wait for.
            strategy: Hybrid strategy configuration.

        Returns:
            The final JobStatus when the target status is reached or job terminates.
        """
        logger.debug(
            "Using hybrid strategy for job %s (websocket timeout: %.1fs)",
            job.id,
            self.websocket_timeout,
        )
        try:
            # Try websocket first with a timeout
            return await asyncio.wait_for(
                WebsocketStrategy.get_status(self, job),
                timeout=self.websocket_timeout,
            )
        except asyncio.TimeoutError:
            # Websocket phase timed out, switch to polling
            logger.debug(
                "Job %s: websocket timeout after %.1fs, switching to polling",
                job.id,
                self.websocket_timeout,
            )
            return await PollingStrategy.get_status(self, job)


class Params(
    CreatorFilter,
    PropertiesFilter,
    PaginationFilter,
    NameFilter,
    JobStatusFilter,
    ProjectRefFilter,
    JobTypeFilter,
    ScopeFilter,
    SortFilter,
    TimeFilter,
):
    """Params for filtering jobs"""


@merge_scope_from_context
@merge_project_from_context
def get_all(
    *,
    name_like: str | None = None,
    name_exact: list[str] | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    job_status: list[JobStatusEnum] | None = None,
    job_type: list[JobType] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> NexusIterator[CompileJobRef | ExecuteJobRef]:
    """Get a NexusIterator over jobs with optional filters."""
    project = project or get_active_project(project_required=False)
    project = cast(ProjectRef, project)

    params = Params(
        name_like=name_like,
        name_exact=name_exact,
        creator_email=creator_email,
        project=project,
        status=(
            JobStatusFilter.convert_status_filters(job_status) if job_status else None
        ),
        job_type=job_type,
        properties=properties,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="Job",
        nexus_url="/api/jobs/v1beta3",
        params=params,
        wrapper_method=_to_jobref,
        nexus_client=get_nexus_client(),
    )


def _to_jobref(data: dict[str, Any]) -> DataframableList[CompileJobRef | ExecuteJobRef]:
    """Parse a json dictionary into a list of JobRefs."""

    job_list: list[CompileJobRef | ExecuteJobRef] = []

    for entry in data["data"]:
        project_id = entry["relationships"]["project"]["data"]["id"]

        project_details = next(
            proj for proj in data["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        system_id: str | None = (
            entry["relationships"]["system"]["data"]["id"]
            if "system" in entry["relationships"]
            else None
        )
        system_details = (
            next(item for item in data["included"] if item["id"] == system_id)
            if system_id is not None
            else None
        )

        system = (
            SystemRef(
                id=UUID(system_id),
                name=system_details["attributes"]["name"],
                provider_name=system_details["attributes"]["provider_name"],
            )
            if system_id is not None and system_details is not None
            else None
        )
        job_type: Type[CompileJobRef] | Type[ExecuteJobRef]
        match entry["attributes"]["job_type"]:
            case JobType.COMPILE:
                job_type = CompileJobRef
            case JobType.EXECUTE:
                job_type = ExecuteJobRef
            case _:
                assert_never(entry["attributes"]["job_type"])

        job_list.append(
            job_type(
                id=entry["id"],
                annotations=Annotations.from_dict(entry["attributes"]),
                job_type=entry["attributes"]["job_type"],
                last_status=JobStatus.from_dict(entry["attributes"]["status"]).status,
                last_message=JobStatus.from_dict(entry["attributes"]["status"]).message,
                last_status_detail=JobStatus.from_dict(entry["attributes"]["status"]),
                project=project,
                system=system,
            )
        )
    return DataframableList(job_list)


@merge_scope_from_context
def get(
    *,
    id: Union[str, UUID, None] = None,
    name: str | None = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    job_status: list[JobStatusEnum] | None = None,
    job_type: list[JobType] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> JobRef:
    """
    Get a single job using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(job_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        name_exact=[name] if name else None,
        creator_email=creator_email,
        project=project,
        properties=properties,
        job_status=job_status,
        job_type=job_type,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_scope_from_context
def _fetch_by_id(
    job_id: UUID | str, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> JobRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/jobs/v1beta3/{job_id}", params=params)

    handle_fetch_errors(res)

    job_data = res.json()

    project_id = job_data["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in job_data["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    system_id: str | None = (
        job_data["data"]["relationships"]["system"]["data"]["id"]
        if "system" in job_data["data"]["relationships"]
        else None
    )

    system_details = (
        next(item for item in job_data["included"] if item["id"] == system_id)
        if system_id is not None
        else None
    )

    system = (
        SystemRef(
            id=UUID(system_id),
            name=system_details["attributes"]["name"],
            provider_name=system_details["attributes"]["provider_name"],
        )
        if system_id is not None and system_details is not None
        else None
    )

    job_type: Type[CompileJobRef] | Type[ExecuteJobRef]
    match job_data["data"]["attributes"]["job_type"]:
        case JobType.COMPILE:
            job_type = CompileJobRef
        case JobType.EXECUTE:
            job_type = ExecuteJobRef
        case _:
            assert_never(job_data["attributes"]["job_type"])

    backend_config_dict = job_data["data"]["attributes"]["definition"]["backend_config"]
    backend_config_class = config_name_to_class[backend_config_dict["type"]]
    backend_config: BackendConfig = backend_config_class(  # type: ignore
        **backend_config_dict
    )

    return job_type(
        id=job_data["data"]["id"],
        annotations=Annotations.from_dict(job_data["data"]["attributes"]),
        job_type=job_data["data"]["attributes"]["job_type"],
        last_status=JobStatus.from_dict(
            job_data["data"]["attributes"]["status"]
        ).status,
        last_message=JobStatus.from_dict(
            job_data["data"]["attributes"]["status"]
        ).message,
        last_status_detail=JobStatus.from_dict(
            job_data["data"]["attributes"]["status"]
        ),
        project=project,
        backend_config_store=backend_config,
        system=system,
    )


def wait_for(
    job: JobRef,
    wait_for_status: JobStatusEnum = JobStatusEnum.COMPLETED,
    timeout: float | None = None,
    strategy: WaitStrategy | None = None,
) -> JobStatus:
    """Check job status until the job is complete (or a specified status).

    Args:
        job: The job to monitor.
        wait_for_status: The status to wait for (default: COMPLETED).
        timeout: Overall timeout in seconds. None for no timeout (default: None).
        strategy: How to monitor the job:
            - WebsocketStrategy(): Real-time updates via websocket.
              Best for short jobs (<10 minutes).
            - PollingStrategy(): Exponential backoff polling.
              Robust for long jobs (>10 minutes).
            - HybridStrategy(): Websocket first, then polling fallback (default).
              Recommended for most use cases.

    Returns:
        The final JobStatus.

    Raises:
        JobError: If the job errors, is cancelled, depleted, or terminated
            (unless that was the status being waited for).
        asyncio.TimeoutError: If the overall timeout is exceeded.

    Examples:
        # Use defaults (hybrid strategy)
        wait_for(job)

        # Custom polling configuration
        wait_for(job, strategy=PollingStrategy(initial_interval=5.0, backoff_factor=1.5))

        # Custom hybrid with polling fallback config
        wait_for(job, strategy=HybridStrategy(
            websocket_timeout=300.0,
            polling=PollingStrategy(max_interval_running=60.0)
        ))
    """
    if strategy is None:
        strategy = HybridStrategy()

    logger.debug(
        "Waiting for job %s with strategy=%s, timeout=%s, target=%s",
        job.id,
        type(strategy).__name__,
        timeout,
        wait_for_status.value,
    )

    coro = strategy.get_status(job)

    if timeout is not None:
        coro = asyncio.wait_for(coro, timeout=timeout)

    job_status = asyncio.run(coro)
    logger.info("Job %s finished with status: %s", job.id, job_status.status.value)

    if (
        job_status.status == JobStatusEnum.ERROR
        and wait_for_status != JobStatusEnum.ERROR
    ):
        raise qnx_exc.JobError(f"Job errored with detail: {job_status.error_detail}")
    if (
        job_status.status == JobStatusEnum.CANCELLED
        and wait_for_status != JobStatusEnum.CANCELLED
    ):
        raise qnx_exc.JobError("Job was cancelled")
    if (
        job_status.status == JobStatusEnum.DEPLETED
        and wait_for_status != JobStatusEnum.DEPLETED
    ):
        raise qnx_exc.JobError("Job has run out of account credits")
    if (
        job_status.status == JobStatusEnum.TERMINATED
        and wait_for_status != JobStatusEnum.TERMINATED
    ):
        raise qnx_exc.JobError("Job has been terminated")

    return job_status


@merge_scope_from_context
def status(job: JobRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER) -> JobStatus:
    """Get the status of a job."""
    resp = get_nexus_client().get(
        f"api/jobs/v1beta3/{job.id}/attributes/status",
        params={"scope": scope.value},
    )
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    job_status = JobStatus.from_dict(resp.json())
    return job_status


@merge_scope_from_context
@overload
def results(
    job: CompileJobRef,
    allow_incomplete: bool = False,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> DataframableList[CompilationResultRef | IncompleteJobItemRef]: ...


@merge_scope_from_context
@overload
def results(
    job: ExecuteJobRef,
    allow_incomplete: bool = False,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> DataframableList[ExecutionResultRef | IncompleteJobItemRef]: ...


@merge_scope_from_context
def results(
    job: CompileJobRef | ExecuteJobRef,
    allow_incomplete: bool = False,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> (
    DataframableList[CompilationResultRef | IncompleteJobItemRef]
    | DataframableList[ExecutionResultRef | IncompleteJobItemRef]
):
    """Get the ResultRefs from a JobRef, if the job is complete.
    To enable fetching results from Jobs with incomplete items, set allow_incomplete=True.
    """
    match job:
        case CompileJobRef():
            return _compile._results(job, allow_incomplete, scope)
        case ExecuteJobRef():
            return _execute._results(job, allow_incomplete, scope)
        case _:
            assert_never(job.job_type)


def retry_submission(
    job: JobRef,
    retry_status: list[JobStatusEnum] | None = None,
    remote_retry_strategy: RemoteRetryStrategy = RemoteRetryStrategy.FULL_RESTART,
    user_group: str | None = None,
) -> None:
    """Retry a job in Nexus according to status(es) or retry strategy.

    By default, jobs with the ERROR status will be retried.
    """
    body: dict[str, str | list[str]] = {"remote_retry_strategy": remote_retry_strategy}

    if user_group is not None:
        body["user_group"] = user_group

    if retry_status is not None:
        body["retry_status"] = [status.name for status in retry_status]

    res = get_nexus_client().post(
        f"/api/jobs/v1beta3/{job.id}/rpc/retry",
        json=body,
    )
    if res.status_code != 202:
        res.raise_for_status()


@merge_scope_from_context
def cancel(job: JobRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER) -> None:
    """Attempt cancellation of a job in Nexus.

    If the job has been submitted to a backend, Nexus will request cancellation of the job.
    """
    res = get_nexus_client().post(
        f"/api/jobs/v1beta3/{job.id}/rpc/cancel",
        json={},
        params={"scope": scope.value},
    )

    if res.status_code != 202:
        res.raise_for_status()


@merge_scope_from_context
def delete(job: JobRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER) -> None:
    """Delete a job in Nexus."""
    res = get_nexus_client().delete(
        f"/api/jobs/v1beta3/{job.id}",
        params={"scope": scope.value},
    )

    if res.status_code != 204:
        res.raise_for_status()


@accept_circuits_for_programs
@merge_properties_from_context
def compile(
    programs: Union[CircuitRef, list[CircuitRef]],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    optimisation_level: int = 2,
    credential_name: str | None = None,
    user_group: str | None = None,
    hypertket_config: HyperTketConfig | None = None,
    timeout: float | None = 300.0,
) -> DataframableList[CircuitRef]:
    """
    Utility method to run a compile job on a program or programs and return a
    DataframableList of the compiled programs.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    compile_job_ref = _compile.start_compile_job(
        programs=programs,
        backend_config=backend_config,
        name=name,
        description=description,
        project=project,
        properties=properties,
        optimisation_level=optimisation_level,
        credential_name=credential_name,
        user_group=user_group,
        hypertket_config=hypertket_config,
    )

    wait_for(job=compile_job_ref, timeout=timeout)

    compile_results = results(compile_job_ref)

    compiled_circuits: list[CircuitRef] = []
    for compile_result in compile_results:
        if isinstance(compile_result, CompilationResultRef):
            compiled_circuits.append(compile_result.get_output())
        elif isinstance(compile_result, IncompleteJobItemRef):
            raise qnx_exc.ResourceFetchFailed(
                f"Compile job item {compile_result.job_item_integer_id} is in status {compile_result.last_status}"
            )
        else:
            assert_never(compile_result)

    return DataframableList(compiled_circuits)


@accept_circuits_for_programs
@merge_properties_from_context
def execute(
    programs: Union[ExecutionProgram, list[ExecutionProgram]],
    n_shots: list[int] | list[None],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    properties: PropertiesDict | None = None,
    project: ProjectRef | None = None,
    valid_check: bool = True,
    wasm_module: WasmModuleRef | None = None,
    gpu_decoder_config: GpuDecoderConfigRef | None = None,
    language: Language = Language.AUTO,
    credential_name: str | None = None,
    user_group: str | None = None,
    timeout: float | None = 300.0,
    max_cost: float | list[float] | list[None] = list(),
    n_qubits: int | list[int] | list[None] = list(),
) -> list[ExecutionResult]:
    """
    Utility method to run an execute job and return the results. Blocks until
    the results are available. See ``qnexus.start_execute_job`` for a function
    that submits the job and returns immediately, rather than waiting for
    results.
    """

    execute_job_ref = _execute.start_execute_job(
        programs=programs,
        n_shots=n_shots,
        backend_config=backend_config,
        name=name,
        description=description,
        properties=properties,
        project=project,
        valid_check=valid_check,
        wasm_module=wasm_module,
        gpu_decoder_config=gpu_decoder_config,
        language=language,
        credential_name=credential_name,
        user_group=user_group,
        max_cost=max_cost,
        n_qubits=n_qubits,
    )

    wait_for(job=execute_job_ref, timeout=timeout)

    execute_results = results(execute_job_ref)

    ex_results: list[ExecutionResult] = []
    for result in execute_results:
        if isinstance(result, ExecutionResultRef):
            ex_results.append(result.download_result())
        elif isinstance(result, IncompleteJobItemRef):
            raise qnx_exc.ResourceFetchFailed(
                f"Compile job item {result.job_item_integer_id} is in status {result.last_status}"
            )
        else:
            assert_never(result)

    return ex_results


@merge_scope_from_context
def cost(
    job: CompileJobRef | ExecuteJobRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> float:
    """Get the HQC cost of a job from a JobRef."""

    resp = get_nexus_client().get(
        f"/api/jobs/v1beta3/{job.id}",
        params={"scope": scope.value},
    )
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    resp_data = resp.json()["data"]
    job_status = resp_data["attributes"]["status"]
    cost = job_status.get("cost")
    return float(cost) if cost is not None else 0.0


@merge_scope_from_context
def cost_confidence(
    job: CompileJobRef | ExecuteJobRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> list[tuple[float, float]]:
    """
    Get the HQC cost and confidence of a job from a JobRef.

    Returns a list of tuples of (cost, confidence) for each job item.
    """
    resp = get_nexus_client().get(
        f"/api/jobs/v1beta3/{job.id}", params={"scope": scope.value}
    )
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )

    # Loop over all items from the API and gather cost_confidence_items
    resp_data = resp.json()["data"]
    items = resp_data["attributes"]["definition"]["items"]
    cost_confidence_items = []
    for i, item in enumerate(items):
        cost_confidence_items.append(
            (
                float(item["status"].get("cost", 0.0)),
                float(item["status"].get("cost_confidence", -1.0)),
            )
        )

    return cost_confidence_items



"""Client API for authentication in Nexus."""

import datetime
import getpass
import time
import warnings
import webbrowser
from http import HTTPStatus

import httpx
import jwt
from colorama import Fore
from pydantic import EmailStr
from rich.console import Console
from rich.panel import Panel

import qnexus.exceptions as qnx_exc
from qnexus.client import (
    VERSION,
    VERSION_HEADER,
    _check_version_headers,
    get_nexus_client,
)
from qnexus.client.utils import consolidate_error, read_token, remove_token, write_token
from qnexus.config import CONFIG

console = Console()


def is_logged_in() -> bool:
    """Check if the user is already logged in by verifying tokens and
    attempting a lightweight authenticated request."""

    try:
        refresh_token = read_token("refresh_token")
        access_token = read_token("access_token")

        # Check that tokens are present
        if not refresh_token or not access_token:
            return False
    except FileNotFoundError:
        return False
    # Check expiry of refresh token (assume JWT)
    try:
        payload = jwt.decode(refresh_token, options={"verify_signature": False})
        exp = payload.get("exp")
        if exp:
            expiry_dt = datetime.datetime.fromtimestamp(exp)
            now = datetime.datetime.now()
            hours_left = (expiry_dt - now).total_seconds() / 3600
            if hours_left < 24:
                msg = (
                    f"Your refresh token expires in less than 24 hours (expires at {expiry_dt}). "
                    "You will need to login again after this time or use force=True to refresh now."
                )
                warnings.warn(msg, category=UserWarning)
    except jwt.PyJWTError:
        pass
    # Try a lightweight authenticated request to check validity
    try:
        client = get_nexus_client()
        resp = client.get("/api/users/v1beta2/me")
        if resp.status_code == HTTPStatus.OK:
            return True
    except (httpx.HTTPError, qnx_exc.AuthenticationError):
        pass
    return False


def _get_auth_client() -> httpx.Client:
    """Getter function for the Nexus auth client."""
    return httpx.Client(
        base_url=f"{CONFIG.url}/auth",
        timeout=None,
        verify=CONFIG.httpx_verify,
    )


def login(force: bool = False) -> None:
    """
    Log in to Quantinuum Nexus using the web browser.

    (if web browser can't be launched, displays the link)
    """
    if not force and is_logged_in():
        print("Already logged in. Tokens are valid.")
        return

    res = _get_auth_client().post(
        "/device/device_authorization",
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        data={"client_id": "scales", "scope": "myqos"},
    )

    user_code = res.json()["user_code"]
    device_code = res.json()["device_code"]
    verification_uri_complete = res.json()["verification_uri_complete"]
    expires_in = res.json()["expires_in"]
    poll_interval = res.json()["interval"]

    webbrowser.open(verification_uri_complete, new=2)

    token_request_body = {
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        "device_code": device_code,
        "client_id": "scales",
    }

    print("🌐 Browser log in initiated.")

    console.print(
        Panel(
            f"""
        Confirm that the browser shows the following code and click 'allow device':

                                     {user_code}
        """,
            width=90,
        )
    )

    print(
        "Browser didn't open automatically? Use this link: "
        f"{Fore.BLUE + verification_uri_complete}"
    )

    polling_for_seconds = 0
    while polling_for_seconds < expires_in:
        time.sleep(poll_interval)
        polling_for_seconds += poll_interval
        resp = _get_auth_client().post(
            "/device/token",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                VERSION_HEADER: VERSION,
            },
            data=token_request_body,
        )
        if (
            resp.status_code == HTTPStatus.BAD_REQUEST
            and resp.json().get("error") == "AUTHORIZATION_PENDING"
        ):
            continue
        if resp.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            continue
        if resp.status_code == HTTPStatus.OK:
            resp_json = resp.json()
            write_token("refresh_token", resp_json["refresh_token"])
            write_token(
                "access_token",
                resp_json["access_token"],
            )
            get_nexus_client(reload=True)
            # spinner.stop()
            print(
                f"✅ Successfully logged in as {resp_json['email']} using the browser."
            )
            _check_version_headers(resp)
            return
        # Fail for all other statuses
        consolidate_error(res=resp, description="Browser Login")
        # spinner.stop()

        return
    raise qnx_exc.AuthenticationError("Browser login Failed, code has expired.")


def login_with_credentials(force: bool = False) -> None:
    """Log in to Nexus using a username and password."""
    if not force and is_logged_in():
        print("Already logged in. Tokens are valid.")
        return
    user_name = input("Enter your Nexus email: ")
    pwd = getpass.getpass(prompt="Enter your Nexus password: ")

    _request_tokens(user=user_name, pwd=pwd)

    print(f"✅ Successfully logged in as {user_name}.")


def login_no_interaction(user: EmailStr, pwd: str, force: bool = False) -> None:
    """Log in to Nexus using a username and password.
    Please be careful with storing credentials in plain text or source code.
    """
    if not force and is_logged_in():
        print("Already logged in. Tokens are valid.")
        return
    _request_tokens(user=user, pwd=pwd)
    print(f"✅ Successfully logged in as {user}.")


def logout() -> None:
    """Clear tokens from file system and the client."""
    remove_token("refresh_token")
    remove_token("access_token")
    get_nexus_client(reload=True)
    print("Successfully logged out.")


def _request_tokens(user: EmailStr, pwd: str) -> None:
    """Method to send login request to Nexus auth api and save tokens."""
    body = {"email": user, "password": pwd}
    try:
        resp = _get_auth_client().post(
            "/login",
            json=body,
            headers={VERSION_HEADER: VERSION},
        )

        mfa_redirect_uri = resp.json().get("redirect_uri", "")
        if mfa_redirect_uri.startswith("/auth/mfa_challenge/"):
            mfa_code = input("Enter your MFA verification code: ")
            body["code"] = mfa_code
            body.pop("password")
            resp = _get_auth_client().post(
                "/mfa_challenge",
                json=body,
            )

        terms_redirect_uri = resp.json().get("redirect_uri", "")
        if terms_redirect_uri.startswith("/auth/terms_challenge"):
            message = "Terms and conditions not accepted. To continue, "
            message += "please accept our new terms and conditions by signing in "
            message += "to the Nexus website https://nexus.quantinuum.com/auth/login."

            # logger.error(message)
            raise qnx_exc.AuthenticationError(message)

        _response_check(resp, "Login")

        myqos_oat = resp.cookies.get("myqos_oat", None)
        myqos_id = resp.cookies.get("myqos_id", None)

        if not myqos_oat or not myqos_id:
            raise qnx_exc.AuthenticationError(
                "Authorization cookies missing from response."
            )

        write_token("refresh_token", myqos_oat)
        write_token("access_token", myqos_id)
        get_nexus_client(reload=True)

        _check_version_headers(resp)

    finally:
        del user
        del pwd
        del body


def _response_check(res: httpx.Response, description: str) -> None:
    """Consolidate as much error-checking of response"""
    # check if token has expired or is generally unauthorized
    resp_json = res.json()
    if res.status_code == HTTPStatus.UNAUTHORIZED:
        raise qnx_exc.AuthenticationError(
            (
                f"Authorization failure attempting: {description}."
                f"\n\nServer Response: {resp_json}"
            )
        )
    if res.status_code != HTTPStatus.OK:
        raise qnx_exc.AuthenticationError(
            f"HTTP error attempting: {description}.\n\nServer Response: {resp_json}"
        )



"""Tests related to running jobs against QSys devices."""

from typing import Callable, ContextManager, cast

import pytest
from hugr.package import Package
from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from quantinuum_schemas.models.backend_config import (
    BackendConfig,
    HeliosConfig,
    HeliosEmulatorConfig,
    SeleneConfig,
)

import qnexus as qnx
from qnexus.models.references import (
    ExecutionResultRef,
    HUGRRef,
    ProjectRef,
    ResultVersions,
)


@pytest.mark.parametrize(
    "backend_config",
    [
        SeleneConfig(),
        HeliosConfig(
            system_name="Helios-1E-lite",
            emulator_config=HeliosEmulatorConfig(),
        ),
    ],
    ids=["SeleneConfig", "Helios-1E-lite"],
)
def test_guppy_execution(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
    backend_config: BackendConfig,
    qa_hugr_package: Package,
) -> None:
    """Test the execution of a guppy program
    on a next-generation QSys device."""

    with create_project(f"project for {test_case_name}") as project_ref:
        n_shots = 10

        hugr_ref = qnx.hugr.upload(
            hugr_package=qa_hugr_package,
            name=f"hugr for {test_case_name}",
            project=project_ref,
        )

        job_ref = qnx.start_execute_job(
            programs=[hugr_ref],
            n_shots=[n_shots],
            backend_config=backend_config,
            project=project_ref,
            name=f"selene job for {test_case_name}",
            n_qubits=[5],
            max_cost=[10.0],
        )

        qnx.jobs.wait_for(job_ref)

        results = qnx.jobs.results(job_ref)

        assert len(results) == 1
        result_ref = results[0]

        assert isinstance(result_ref, ExecutionResultRef)
        assert isinstance(result_ref.download_backend_info(), BackendInfo)
        assert isinstance(result_ref.get_input(), HUGRRef)

        assert result_ref.get_input().id == hugr_ref.id

        qsys_result = cast(QsysResult, result_ref.download_result())
        assert len(qsys_result.results) == n_shots
        assert qsys_result.results[0].entries[0][0] == "teleported"
        assert qsys_result.results[0].entries[0][1] == 1

        # check some QsysResults functionality
        assert len(qsys_result.collated_counts().items()) > 0

        # Assert we can get the same result for v4 results
        v4_qsys_result = cast(
            QsysResult, result_ref.download_result(version=ResultVersions.RAW)
        )
        assert len(v4_qsys_result.results) == n_shots
        assert v4_qsys_result.results[0].entries[0][0] == "USER:BOOL:teleported"
        assert v4_qsys_result.results[0].entries[0][1] == 1

        # This doesn't seem to work with v4 results, thinks everything is going to be a bit.
        # assert len(qsys_result.collated_counts().items()) > 0


def test_hugr_costing(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
    qa_hugr_package: Package,
) -> None:
    """Test the costing of a Hugr program on a cost checking device."""

    with create_project(f"project for {test_case_name}") as project_ref:
        hugr_ref = qnx.hugr.upload(
            hugr_package=qa_hugr_package,
            name=f"hugr for {test_case_name}",
            project=project_ref,
        )

        # Check that we can get a cost estimate (using Helios-1SC)
        cost = qnx.hugr.cost(
            programs=[hugr_ref],
            n_shots=[10],
            project=project_ref,
        )
        assert isinstance(cost, float)


def test_hugr_cost_confidence(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
    qa_hugr_package: Package,
) -> None:
    """Test the cost confidence of a Hugr program on a cost checking device."""

    with create_project(f"project for {test_case_name}") as project_ref:
        hugr_ref = qnx.hugr.upload(
            hugr_package=qa_hugr_package,
            name=f"hugr for {test_case_name}",
            project=project_ref,
        )

        # Check that we can get a cost confidence estimate (using Helios-1SC)
        cost_confidence = qnx.hugr.cost_confidence(
            programs=[hugr_ref],
            n_shots=[10],
            project=project_ref,
        )
        assert isinstance(cost_confidence, list)
        assert len(cost_confidence) > 0



"""Client API for QIR in Nexus."""

import base64
import warnings
from datetime import datetime
from typing import Any, Literal, Union, cast
from uuid import UUID

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
    merge_scope_from_context,
)
from qnexus.models import QuantinuumConfig
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    NameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import (
    DataframableList,
    ExecutionProgram,
    ProjectRef,
    QIRRef,
)
from qnexus.models.scope import ScopeFilterEnum


class Params(
    SortFilter,
    PaginationFilter,
    NameFilter,
    CreatorFilter,
    ProjectRefFilter,
    PropertiesFilter,
    TimeFilter,
    ScopeFilter,
):
    """Params for filtering QIRs."""


@merge_scope_from_context
@merge_project_from_context
def get_all(
    *,
    name_like: str | None = None,
    name_exact: list[str] | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> NexusIterator[QIRRef]:
    """Get a NexusIterator over QIRs with optional filters."""

    params = Params(
        name_like=name_like,
        name_exact=name_exact,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="QIR",
        nexus_url="/api/qir/v1beta",
        params=params,
        wrapper_method=_to_qir_ref,
        nexus_client=get_nexus_client(),
    )


def _to_qir_ref(page_json: dict[str, Any]) -> DataframableList[QIRRef]:
    """Convert JSON response dict to a list of QIRRefs."""

    qir_refs: DataframableList[QIRRef] = DataframableList([])

    for qir_data in page_json["data"]:
        project_id = qir_data["relationships"]["project"]["data"]["id"]
        project_details = next(
            proj for proj in page_json["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        qir_refs.append(
            QIRRef(
                id=UUID(qir_data["id"]),
                annotations=Annotations.from_dict(qir_data["attributes"]),
                project=project,
            )
        )
    return qir_refs


@merge_scope_from_context
def get(
    *,
    id: Union[UUID, str, None] = None,
    name: str | None = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum = ScopeFilterEnum.USER,
) -> QIRRef:
    """
    Get a single QIR using filters. Throws an exception if the filters do
    not match exactly one object.
    """

    if id:
        return _fetch_by_id(qir_id=id, scope=scope)

    return get_all(
        name_exact=[name] if name else None,
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_properties_from_context
def upload(
    qir: bytes,
    name: str,
    project: ProjectRef | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> QIRRef:
    """Upload a QIR to Nexus."""
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    attributes = {"contents": _encode_qir(qir)}

    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes.update(annotations)
    relationships = {"project": {"data": {"id": str(project.id), "type": "project"}}}

    req_dict = {
        "data": {
            "attributes": attributes,
            "relationships": relationships,
            "type": "qir",
        }
    }

    res = get_nexus_client().post("/api/qir/v1beta", json=req_dict)

    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return QIRRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        project=project,
    )


@merge_properties_from_context
def update(
    ref: QIRRef,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> QIRRef:
    """Update the annotations on a QIRRef."""
    ref_annotations = ref.annotations.model_dump()
    annotations = Annotations(
        name=name,
        description=description,
        properties=properties if properties else PropertiesDict(),
    ).model_dump(exclude_none=True)
    ref_annotations.update(annotations)

    req_dict = {
        "data": {
            "attributes": annotations,
            "relationships": {},
            "type": "qir",
        }
    }

    res = get_nexus_client().patch(f"/api/qir/v1beta/{ref.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_dict = res.json()["data"]

    return QIRRef(
        id=UUID(res_dict["id"]),
        annotations=Annotations.from_dict(res_dict["attributes"]),
        project=ref.project,
    )


def cost(
    programs: QIRRef | list[QIRRef],
    n_shots: int | list[int],
    project: ProjectRef | None = None,
    system_name: Literal["Helios-1"] = "Helios-1",
) -> float:
    """Estimate the cost (in HQC) of running QIR programs for n_shots
    number of shots on a Quantinuum Helios system.

    NB: This will execute a costing job on a dedicated cost estimation device.
        Once run, the cost will be visible also in the Nexus web portal
        as part of the job.
    """
    import qnexus as qnx

    warnings.warn(
        "qir.cost() is deprecated. Please update to use qir.cost_confidence() instead.",
        category=DeprecationWarning,
    )

    if isinstance(programs, QIRRef):
        programs = [programs]

    job_ref = qnx.start_execute_job(
        programs=cast(list[ExecutionProgram], programs),
        n_shots=n_shots,
        backend_config=QuantinuumConfig(device_name=f"{system_name}SC"),
        project=project,
        name="QIR cost estimation job",
    )
    status = qnx.jobs.wait_for(job_ref)

    return cast(float, status.cost)


def cost_confidence(
    programs: QIRRef | list[QIRRef],
    n_shots: int | list[int],
    project: ProjectRef | None = None,
    system_name: Literal["Helios-1"] = "Helios-1",
) -> list[tuple[float, float]]:
    """Estimate the cost (in HQC) of running QIR programs for n_shots
    number of shots on a Quantinuum Helios system.

    Returns a list of tuples of (cost, confidence) for each job item.

    NB: This will execute a costing job on a dedicated cost estimation device.
        Once run, the cost will be visible also in the Nexus web portal
        as part of the job.
    """
    import qnexus as qnx

    if isinstance(programs, QIRRef):
        programs = [programs]

    job_ref = qnx.start_execute_job(
        programs=cast(list[ExecutionProgram], programs),
        n_shots=n_shots,
        backend_config=QuantinuumConfig(device_name=f"{system_name}SC"),
        project=project,
        name="QIR cost estimation job",
    )

    qnx.jobs.wait_for(job_ref)

    return qnx.jobs.cost_confidence(job_ref)


@merge_scope_from_context
def _fetch_by_id(
    qir_id: UUID | str, scope: ScopeFilterEnum = ScopeFilterEnum.USER
) -> QIRRef:
    """Utility method for fetching directly by a unique identifier."""

    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/qir/v1beta/{qir_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    project_id = res_dict["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in res_dict["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    return QIRRef(
        id=UUID(res_dict["data"]["id"]),
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        project=project,
    )


@merge_scope_from_context
def _fetch_qir(handle: QIRRef, scope: ScopeFilterEnum = ScopeFilterEnum.USER) -> bytes:
    """Utility method for fetching QIR bytes from a QIRRef."""
    res = get_nexus_client().get(
        f"/api/qir/v1beta/{handle.id}",
        params={"scope": scope.value},
    )
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    contents: str = res.json()["data"]["attributes"]["contents"]
    return _decode_qir(contents)


def _encode_qir(qir: bytes) -> str:
    """Utility method for encoding QIR bytes as base64-encoded string"""
    return base64.b64encode(qir).decode("utf-8")


def _decode_qir(contents: str) -> bytes:
    """Utility method for decoding a base64-encoded string into QIR bytes"""
    return base64.b64decode(contents)



"""Definitions of reference proxy objects to data in Nexus."""

from __future__ import annotations

from abc import abstractmethod
from copy import copy
from enum import Enum
from typing import (
    Annotated,
    Any,
    Iterable,
    Literal,
    Optional,
    Protocol,
    TypeAlias,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

import pandas as pd
from hugr.package import Package
from hugr.qsystem.result import QsysResult
from pydantic import ConfigDict, Field
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Circuit
from pytket.wasm.wasm import WasmModuleHandler
from quantinuum_schemas.models.backend_config import BackendConfig

from qnexus.context import merge_scope_from_context
from qnexus.exceptions import IncompatibleResultVersion
from qnexus.models.annotations import Annotations
from qnexus.models.job_status import JobStatus, JobStatusEnum
from qnexus.models.references.base import BaseRef
from qnexus.models.references.projects import ProjectRef
from qnexus.models.scope import ScopeFilterEnum
from qnexus.models.utils import assert_never

__all__ = [
    "BaseRef",  # re-export
    "CircuitRef",
    "CompilationPassRef",
    "CompilationResultRef",
    "CompileJobRef",
    "Dataframable",
    "DataframableList",
    "deserialize_nexus_ref",
    "ExecuteJobRef",
    "ExecutionProgram",
    "ExecutionResult",
    "ExecutionResultRef",
    "GpuDecoderConfigRef",
    "HUGRRef",
    "IncompleteJobItemRef",
    "JobRef",
    "JobType",
    "JobType",
    "ProjectRef",  # re-export
    "QIRRef",
    "QIRResult",
    "ref_name_to_class",
    "Ref",
    "ResultType",
    "ResultVersions",
    "SystemRef",
    "TeamRef",
    "UserRef",
    "WasmModuleRef",
]


class Dataframable(Protocol):
    """Protocol for structural subtyping of classes that
    have a default representation as a pandas.DataFrame."""

    @abstractmethod
    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        raise NotImplementedError


T = TypeVar("T", bound=Dataframable)


class DataframableList(list[T]):
    """A Python list that implements the Dataframable protocol."""

    def __init__(self, iterable: Iterable[T]) -> None:
        super().__init__(item for item in iterable)

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        if len(self) == 0:
            return pd.DataFrame()
        return pd.concat([item.df() for item in self], ignore_index=True)


class TeamRef(BaseRef):
    """Proxy object to a Team in Nexus."""

    name: str
    description: Optional[str]
    id: UUID
    type: Literal["TeamRef"] = "TeamRef"

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "description": self.description,
                "id": self.id,
            },
            index=[0],
        )


class UserRef(BaseRef):
    """Proxy object to a User in Nexus."""

    display_name: Optional[str]
    id: UUID
    type: Literal["UserRef"] = "UserRef"

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.display_name,
                "id": self.id,
            },
            index=[0],
        )


class SystemRef(BaseRef):
    """Proxy object to a System in Nexus."""

    id: UUID
    name: str
    provider_name: str
    type: Literal["SystemRef"] = "SystemRef"

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "id": self.id,
                "name": self.name,
                "provider_name": self.provider_name,
            },
            index=[0],
        )


class CircuitRef(BaseRef):
    """Proxy object to a Circuit in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _circuit: Circuit | None = None
    type: Literal["CircuitRef"] = "CircuitRef"

    @merge_scope_from_context
    def download_circuit(
        self, scope: ScopeFilterEnum = ScopeFilterEnum.USER
    ) -> Circuit:
        """Get a copy of the circuit as a pytket ``Circuit`` object."""
        if self._circuit:
            return self._circuit.copy()
        from qnexus.client.circuits import _fetch_circuit

        self._circuit = _fetch_circuit(self, scope=scope)
        return self._circuit.copy()

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class WasmModuleRef(BaseRef):
    """Proxy object to a WasmModule in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: WasmModuleHandler | None = None
    type: Literal["WasmModuleRef"] = "WasmModuleRef"

    def download_wasm_contents(self) -> WasmModuleHandler:
        """Get the contents of the original uploaded WASM."""
        if self._contents:
            return self._contents

        from qnexus.client.wasm_modules import _fetch_wasm_module

        self._contents = _fetch_wasm_module(self)
        return self._contents

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class GpuDecoderConfigRef(BaseRef):
    """Proxy object to a GpuDecoderConfig in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: str | None = None
    type: Literal["GpuDecoderConfigRef"] = "GpuDecoderConfigRef"

    def download_gpu_decoder_config_contents(self) -> str:
        """Get the contents of the original uploaded gpu decoder config."""
        if self._contents:
            return self._contents

        from qnexus.client.gpu_decoder_configs import (
            _fetch_gpu_decoder_config,
        )

        self._contents = _fetch_gpu_decoder_config(self)
        return self._contents

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class HUGRRef(BaseRef):
    """Proxy object to a HUGR in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: Package | None = None
    _bytes: bytes | None = None
    type: Literal["HUGRRef"] = "HUGRRef"

    def download_hugr(self) -> Package:
        """Get the HUGR Package of the original uploaded HUGR."""

        if self._contents:
            return self._contents

        from qnexus.client.hugr import _fetch_hugr_package

        self._contents = _fetch_hugr_package(self)
        return self._contents

    def download_hugr_bytes(self) -> bytes:
        """Get the HUGR bytes of the original uploaded HUGR."""

        if self._bytes:
            return self._bytes

        from qnexus.client.hugr import _fetch_hugr_bytes

        self._bytes = _fetch_hugr_bytes(self)
        return self._bytes

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class QIRRef(BaseRef):
    """Proxy object to a QIR program in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: bytes | None = None
    type: Literal["QIRRef"] = "QIRRef"

    def download_qir(self) -> bytes:
        """Get the QIR program."""

        if self._contents:
            return self._contents

        from qnexus.client.qir import _fetch_qir

        self._contents = _fetch_qir(self)
        return self._contents

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class JobType(str, Enum):
    """Enum for a job's type."""

    EXECUTE = "execute"
    COMPILE = "compile"


class JobRef(BaseRef):
    """Proxy object to a Job in Nexus."""

    model_config = ConfigDict(frozen=False)

    annotations: Annotations
    job_type: JobType
    last_status: JobStatusEnum
    last_message: str
    last_status_detail: JobStatus | None = None
    project: ProjectRef
    system: SystemRef | None = None
    id: UUID
    backend_config_store: BackendConfig | None = None
    type: Literal["JobRef", "CompileJobRef", "ExecuteJobRef"] = "JobRef"

    @property
    def backend_config(self) -> BackendConfig:
        """Fetch the backend_config for a job."""
        from qnexus.client.jobs import _fetch_by_id

        if self.backend_config_store:
            return self.backend_config_store
        self.backend_config_store = cast(
            BackendConfig, _fetch_by_id(self.id).backend_config_store
        )
        return self.backend_config_store

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "job_type": self.job_type,
                    "last_status": self.last_status,
                    "project": self.project.annotations.name,
                    "backend_config": self.backend_config.__class__.__name__,
                    "system": self.system.name if self.system else "Unknown",
                    "cost": (
                        self.last_status_detail.cost
                        if self.last_status_detail
                        else "Unknown"
                    ),
                    "id": self.id,
                },
                index=[0],
            )
        )


class CompileJobRef(JobRef, BaseRef):
    """Proxy object to a CompileJob in Nexus."""

    model_config = ConfigDict(frozen=False)

    job_type: JobType = JobType.COMPILE
    type: Literal["CompileJobRef"] = "CompileJobRef"


class ExecuteJobRef(JobRef, BaseRef):
    """Proxy object to an ExecuteJob in Nexus."""

    model_config = ConfigDict(frozen=False)

    job_type: JobType = JobType.EXECUTE
    type: Literal["ExecuteJobRef"] = "ExecuteJobRef"


class CompilationResultRef(BaseRef):
    """Proxy object to the results of a circuit compilation in Nexus."""

    annotations: Annotations
    project: ProjectRef
    last_status_detail: JobStatus | None = None
    _input_circuit: CircuitRef | None = None
    _output_circuit: CircuitRef | None = None
    _compilation_passes: DataframableList[CompilationPassRef] | None = None
    id: UUID  # compilation id
    job_item_id: UUID | None = None
    job_item_integer_id: int | None = None
    type: Literal["CompilationResultRef"] = "CompilationResultRef"

    def get_input(self) -> CircuitRef:
        """Get the CircuitRef of the original circuit."""
        if self._input_circuit:
            return self._input_circuit

        from qnexus.client.jobs._compile import _fetch_compilation_output

        (self._input_circuit, self._output_circuit) = _fetch_compilation_output(self)
        return self._input_circuit

    def get_output(self) -> CircuitRef:
        """Get the CircuitRef of the compiled circuit."""
        if self._output_circuit:
            return self._output_circuit

        from qnexus.client.jobs._compile import _fetch_compilation_output

        (self._input_circuit, self._output_circuit) = _fetch_compilation_output(self)
        return self._output_circuit

    def get_passes(self) -> DataframableList[CompilationPassRef]:
        """Get information on the compilation passes and the output circuits (if available)."""
        if self._compilation_passes:
            return copy(self._compilation_passes)

        self._compilation_passes = self._get_compile_results()
        return copy(self._compilation_passes)

    def _get_compile_results(
        self,
    ) -> DataframableList[CompilationPassRef]:
        """Utility method to retrieve the passes and output circuit."""
        from qnexus.client.jobs._compile import _fetch_compilation_passes

        passes = _fetch_compilation_passes(self)
        return passes

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                    "job_item_id": self.job_item_id,
                    "job_item_integer_id": self.job_item_integer_id,
                },
                index=[0],
            )
        )


class ResultType(str, Enum):
    """Enum for a results's type."""

    PYTKET = "PYTKET"
    QSYS = "QSYS"


class ProgramType(str, Enum):
    """Enum for a program's type."""

    CIRCUIT = "circuit"
    HUGR = "hugr"
    QIR = "qir"


class ResultVersions(int, Enum):
    """Enumerate the valid values for requesting results in a specific format"""

    DEFAULT = 3
    RAW = 4


class QIRResult:
    results: str

    def __init__(self, results: str):
        self.results = results


ExecutionProgram: TypeAlias = CircuitRef | HUGRRef | QIRRef
ExecutionResult: TypeAlias = QsysResult | BackendResult | QIRResult


class ExecutionResultRef(BaseRef):
    """Proxy object to the results of a circuit execution through Nexus."""

    annotations: Annotations
    project: ProjectRef
    result_type: ResultType = ResultType.PYTKET
    cost: float | None = None
    last_status_detail: JobStatus | None = None
    _input_program: ExecutionProgram | None = None
    _result: ExecutionResult | None = None
    _result_version: ResultVersions | None = None
    _backend_info: BackendInfo | None = None
    id: UUID
    job_item_id: UUID | None = None
    job_item_integer_id: int | None = None
    type: Literal["ExecutionResultRef"] = "ExecutionResultRef"

    def get_input(self) -> ExecutionProgram:
        """Get the Program Ref of the input program."""
        if self._input_program:
            return self._input_program

        (
            self._result,
            self._backend_info,
            self._input_program,
        ) = self._get_execute_results(ResultVersions.DEFAULT)
        self._result_version = ResultVersions.DEFAULT
        return copy(self._input_program)

    def download_result(
        self, version: ResultVersions = ResultVersions.DEFAULT
    ) -> ExecutionResult:
        """Get a copy of the result of the program execution."""
        if self._result and self._result_version == version:
            return copy(self._result)

        (
            self._result,
            self._backend_info,
            self._input_program,
        ) = self._get_execute_results(version=version)
        self._result_version = version
        return copy(self._result)

    def download_backend_info(self) -> BackendInfo:
        """Get a copy of the pytket BackendInfo."""
        if self._backend_info:
            return copy(self._backend_info)

        (
            self._result,
            self._backend_info,
            self._input_program,
        ) = self._get_execute_results(ResultVersions.DEFAULT)
        self._result_version = ResultVersions.DEFAULT
        return copy(self._backend_info)

    def _get_execute_results(
        self, version: ResultVersions
    ) -> tuple[ExecutionResult, BackendInfo, ExecutionProgram]:
        """Utility method to retrieve the passes and output circuit.
        result_version can be passed to request v4 results for qsys results only.

        Default results for any program type on H series devices is pytket style results
        Default result for QIR programs on NG devices is QIR standard compliant results.
        Default result for HUGR programs is NG results.
        """
        from qnexus.client.jobs._execute import (
            _fetch_pytket_execution_result,
            _fetch_qsys_execution_result,
        )

        match self.result_type:
            case ResultType.PYTKET:
                if version != ResultVersions.DEFAULT:
                    raise IncompatibleResultVersion(
                        "pytket results can only be fetched in the default version"
                    )
                return _fetch_pytket_execution_result(self)
            case ResultType.QSYS:
                return _fetch_qsys_execution_result(self, version)
            case _:
                assert_never(self.result_type)

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                    "result_type": self.result_type,
                    "cost": self.cost,
                    "job_item_id": self.job_item_id,
                    "job_item_integer_id": self.job_item_integer_id,
                },
                index=[0],
            )
        )


class IncompleteJobItemRef(BaseRef):
    """Proxy object to a Job Item in Nexus that is not complete."""

    annotations: Annotations
    id: UUID = UUID(int=0)  # Incomplete items have no result ID
    job_item_id: UUID | None = None
    job_item_integer_id: int | None = None
    program_type: ProgramType = ProgramType.CIRCUIT
    program_id: UUID | None = None
    _input_program: ExecutionProgram | None = None
    project: ProjectRef
    job_type: JobType
    last_status: JobStatusEnum
    last_message: str
    last_status_detail: JobStatus | None = None
    type: Literal["IncompleteJobItemRef"] = "IncompleteJobItemRef"

    def get_input(self) -> ExecutionProgram:
        """Get the Program Ref of the input program."""
        if self._input_program:
            return self._input_program

        self._input_program = self._get_input_program()
        return copy(self._input_program)

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                    "last_status": self.last_status,
                    "job_item_id": self.job_item_id,
                    "job_item_integer_id": self.job_item_integer_id,
                },
                index=[0],
            )
        )

    def _get_input_program(
        self,
    ) -> ExecutionProgram:
        """Utility method to retrieve the input program for an
        incomplete job item.
        """
        from qnexus.client.circuits import _fetch_by_id as _fetch_circuit_by_id
        from qnexus.client.hugr import _fetch_by_id as _fetch_hugr_by_id
        from qnexus.client.qir import _fetch_by_id as _fetch_qir_by_id

        if not self.program_id:
            raise ValueError(
                "Cannot fetch input program for incomplete job item without a program_id"
            )

        match self.program_type:
            case ProgramType.CIRCUIT:
                return _fetch_circuit_by_id(self.program_id)
            case ProgramType.QIR:
                return _fetch_qir_by_id(self.program_id)
            case ProgramType.HUGR:
                return _fetch_hugr_by_id(self.program_id)
            case _:
                assert_never(self.program_type)


class CompilationPassRef(BaseRef):
    """Proxy object to a compilation pass that was applied on a circuit in Nexus."""

    pass_name: str
    input_circuit: CircuitRef
    output_circuit: CircuitRef
    id: UUID
    type: Literal["CompilationPassRef"] = "CompilationPassRef"

    def get_input(self) -> CircuitRef:
        """Get the CircuitRef of the original circuit."""
        return self.input_circuit

    def get_output(self) -> CircuitRef:
        """Get the CircuitRef of the compiled circuit."""
        return self.output_circuit

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "pass name": self.pass_name,
                "input": self.input_circuit.annotations.name,
                "output": self.output_circuit.annotations.name,
                "id": self.id,
            },
            index=[0],
        )


Ref = Annotated[
    Union[
        TeamRef,
        UserRef,
        ProjectRef,
        CircuitRef,
        WasmModuleRef,
        GpuDecoderConfigRef,
        HUGRRef,
        QIRRef,
        JobRef,
        CompileJobRef,
        ExecuteJobRef,
        CompilationResultRef,
        ExecutionResultRef,
        CompilationPassRef,
        SystemRef,
        IncompleteJobItemRef,
    ],
    Field(discriminator="type"),
]


ref_name_to_class: dict[str, Ref] = {
    config_type.__name__: config_type  # type: ignore
    for config_type in BaseRef.__subclasses__()
}


def deserialize_nexus_ref(jsonable: dict[str, Any]) -> Ref:
    """Deserialize something that should be a subclass of BaseRef based on
    the value of its 'type' field."""
    ref_type = jsonable["type"]
    if ref_type in ref_name_to_class.keys():
        ref_class = ref_name_to_class[ref_type]
        return ref_class(**jsonable)  # type: ignore
    raise ValueError(
        f"Cannot deserialize as {ref_type}, no known class matches that value."
    )



"""Checks for circuit functionality."""

import uuid
from datetime import datetime
from unittest import mock

import pytest

from qnexus.context import using_scope
from qnexus.models.annotations import Annotations
from qnexus.models.references import CircuitRef, ProjectRef
from qnexus.models.scope import ScopeFilterEnum


@pytest.mark.parametrize(
    "chosen_scope,expected_scope",
    [
        (
            None,
            ScopeFilterEnum.USER,
        ),
        (
            ScopeFilterEnum.USER,
            ScopeFilterEnum.USER,
        ),
        (
            ScopeFilterEnum.HIGHEST,
            ScopeFilterEnum.HIGHEST,
        ),
    ],
)
@mock.patch("qnexus.client.circuits._fetch_circuit")
def test_uses_scope_when_downloading_circuit(
    fetch_circuit: mock.MagicMock,
    chosen_scope: ScopeFilterEnum,
    expected_scope: ScopeFilterEnum,
) -> None:
    """Given an circuit ref, it will hit the Nexus API with the appropriate
    scope when .download_circuit() is called."""

    circuit_ref = CircuitRef(
        id=uuid.uuid4(),
        annotations=Annotations(),
        project=ProjectRef(
            id=uuid.uuid4(), annotations=Annotations(), contents_modified=datetime.now()
        ),
    )
    if chosen_scope:
        with using_scope(chosen_scope):
            circuit_ref.download_circuit()
    else:
        circuit_ref.download_circuit()

    fetch_circuit.assert_called_once_with(circuit_ref, scope=expected_scope)



"""Test basic functionality relating to the hugr module."""

from typing import Callable, ContextManager

from hugr.package import Package

import qnexus as qnx
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import HUGRRef, ProjectRef, Ref


def test_hugr_create_and_update(
    test_case_name: str,
    qa_hugr_package: Package,
    create_property_in_project: Callable[..., ContextManager[ProjectRef]],
) -> None:
    """Test that we can create a hugr and add a property value."""
    project_name = f"project for {test_case_name}"
    property_name = f"property for {test_case_name}"

    with create_property_in_project(
        project_name=project_name,
        property_name=property_name,
        property_type="string",
        required=False,
    ) as my_proj:
        hugr_name = f"hugr for {test_case_name}"
        my_new_hugr = qnx.hugr.upload(
            hugr_package=qa_hugr_package, name=hugr_name, project=my_proj
        )

        assert isinstance(my_new_hugr, HUGRRef)

        test_prop_value = "foo"
        updated_hugr_ref = qnx.hugr.update(
            ref=my_new_hugr,
            properties=PropertiesDict({property_name: test_prop_value}),
        )

        assert updated_hugr_ref.annotations.properties[property_name] == test_prop_value


def test_hugr_download(
    test_case_name: str,
    qa_hugr_package: Package,
    create_hugr_in_project: Callable[[str, str, Package], ContextManager[HUGRRef]],
) -> None:
    """Test that valid HUGR can be extracted from an uploaded HUGR module."""
    project_name = f"project for {test_case_name}"
    hugr_name = f"hugr for {test_case_name}"

    with create_hugr_in_project(
        project_name,
        hugr_name,
        qa_hugr_package,
    ) as hugr_ref:
        # Temporarily disabled due to missing functionality in hugr
        downloaded_hugr_package = hugr_ref.download_hugr()
        assert isinstance(downloaded_hugr_package, Package)
        # Direct equality comparison not supported
        assert qa_hugr_package.to_bytes() == downloaded_hugr_package.to_bytes()

        downloaded_hugr_bytes = hugr_ref.download_hugr_bytes()
        assert isinstance(downloaded_hugr_bytes, bytes)
        assert qa_hugr_package.to_bytes() == downloaded_hugr_bytes


def test_hugr_get_by_id(
    test_case_name: str,
    qa_hugr_package: Package,
    create_hugr_in_project: Callable[[str, str, Package], ContextManager[HUGRRef]],
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can get a HUGRRef by its ID and the HUGRRef
    serialisation round trip."""
    project_name = f"project for {test_case_name}"
    hugr_name = f"hugr for {test_case_name}"

    with create_hugr_in_project(
        project_name,
        hugr_name,
        qa_hugr_package,
    ):
        my_proj = qnx.projects.get(name=project_name)
        my_hugr_ref = qnx.hugr.get(name=hugr_name, project=my_proj)

        hugr_ref_by_id = qnx.hugr.get(id=my_hugr_ref.id)

        assert hugr_ref_by_id == my_hugr_ref

        test_ref_serialisation("hugr", hugr_ref_by_id)


def test_hugr_get_all(
    test_case_name: str,
    qa_hugr_package: Package,
    create_hugr_in_project: Callable[[str, str, Package], ContextManager[HUGRRef]],
) -> None:
    """Test that we can get all HUGRRefs in a project."""
    project_name = f"project for {test_case_name}"
    hugr_name = f"hugr for {test_case_name}"

    with create_hugr_in_project(
        project_name,
        hugr_name,
        qa_hugr_package,
    ):
        my_proj = qnx.projects.get(name=project_name)

        hugrs = qnx.hugr.get_all(project=my_proj)

        assert hugrs.count() >= 1
        assert isinstance(hugrs.list()[0], HUGRRef)
